#include <iostream>
#include <fstream>
#include <vector>
#include <pvm3.h>
using namespace std;

void read_input(const char* file, vector<double>& inp);
void master_main(const vector<double>& inp);
void worker_main();


int main(int argc, char* argv[])
{
    if (pvm_parent() == PvmNoParent)
    {
        if (argc < 2)
        {
            cerr << "Hasznlat: " << argv[0] << " <input-file>" << endl;
            return 1;
        }

        vector<double> input;
        read_input(argv[1], input);

        master_main(input);
    }
    else
        worker_main();
    pvm_exit();
    return 0;
}

void read_input(const char* file, vector<double>& inp)
{
    ifstream f(file);
    if (!f)
    {
        cerr << file << ": olvassi hiba." << endl;
        exit(2);
    }

    double d;
    while (f >> d) inp.push_back(d);
}

void send_data(double data, int tid)
{
    pvm_initsend(PvmDataDefault);
    pvm_pkdouble(&data, 1, 1);
    pvm_send(tid, 1);
}

void recv_data(double& data, int tid)
{
    pvm_recv(tid, 1);
    pvm_upkdouble(&data, 1, 1);
}

void master_main(const vector<double>& inp)
{
    int tasknum = inp.size();
    int* tids = new int[tasknum];
    if (pvm_spawn("osszeg", 0, PvmTaskDefault, 0, tasknum, tids) < tasknum)
    {
        pvm_perror("spawn");
        exit(3);
    }

    for (int i=0; i<tasknum; ++i)
        send_data(inp[i], tids[i]);

    for (int i=0; i<tasknum; ++i)
    {
        double result;
        recv_data(result, tids[i]);
        cout << i+1 << ".: " << result << endl;
    }
}

void worker_main()
{
    int* tids;
    // numtask: a ttelben "n", az adatok szma (megegyezik a taszkok szmval)
    int numtask = pvm_siblings(&tids);

    // index: a taszk az "index" sorszm rtkadst ismtelgeti
    int index = 0;
    while (index < numtask && tids[index] != pvm_mytid())
        ++index;

    // data: mindig a "gs(index,k(index))" rtket tartalmazza
    double data;
    recv_data(data, pvm_parent());

    // gap: mindig a "t(index)" rtket tartalmazza (megegyezik a "2^k(i)"-vel)
    int gap = 1;

    // Az i. rtkadshoz kt taszkra van szksg: az i s az i-t(i) indexre.
    // Az i index trolja a gs(i, k(i))-t s vgzi a szmolst, az i-t(i)
    // index trolja a gs(i-t(i),k(i))-t.
    //
    // Az i. rtkads egyik felttele az i-t(i)>=0. Ennek felel meg az
    // index-gap >= 0, ekkor kell adatot fogadni s szmolni.
    //
    // Ahhoz, hogy ez az rtkads mkdjn, az i-t(i) index taszknak el kell
    // kldenie a gs(i,k(i)) rtket akkor, amikor k(i-t(i))=k(i). Ez a msik
    // felttel, ennek a teljeslst az zenet megrkezse jelzi.
    //
    // A k(i-t(i))=k(i) az i-t(i). taszk szemszgbl (ahol index=i-t(i)) azt
    // jelenti, hogy gs(index,k(index))-t az index+t(index). taszknak kell
    // elkldeni. Ennek a szksgessgt rja le az index+gap < numtask
    // felttel.
    while (index - gap >= 0 || index + gap < numtask)
    {
        if (index + gap < numtask)
            send_data(data, tids[index + gap]);

        if (index - gap >= 0)
        {
            double other;
            recv_data(other, tids[index - gap]);
            
            data += other;
        }

        gap *= 2;
    }

    send_data(data, pvm_parent());
}
